/*
 * Calculate Android Copyright <C> 2020 - 2025 Jonas Lochmann
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation version 3 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see https://www.gnu.org/licenses/
 */
package com.rechnen.app.ui.training

import android.annotation.SuppressLint
import androidx.compose.foundation.layout.Box
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.ColumnScope
import androidx.compose.foundation.layout.Row
import androidx.compose.foundation.layout.RowScope
import androidx.compose.foundation.layout.fillMaxHeight
import androidx.compose.foundation.layout.padding
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.Check
import androidx.compose.material.icons.filled.Close
import androidx.compose.material3.Button
import androidx.compose.material3.Icon
import androidx.compose.material3.MaterialTheme
import androidx.compose.material3.ShapeDefaults
import androidx.compose.material3.Text
import androidx.compose.runtime.Composable
import androidx.compose.ui.BiasAlignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.graphics.vector.ImageVector
import androidx.compose.ui.layout.Layout
import androidx.compose.ui.layout.Measurable
import androidx.compose.ui.layout.MeasurePolicy
import androidx.compose.ui.layout.MeasureResult
import androidx.compose.ui.layout.MeasureScope
import androidx.compose.ui.unit.Constraints
import androidx.compose.ui.unit.dp
import com.rechnen.app.data.modelparts.InputConfiguration

@Composable
@SuppressLint("UnusedBoxWithConstraintsScope")
fun KeyboardView(
    configuration: InputConfiguration,
    confirmButton: ConfirmButton,
    enabled: Boolean,
    listener: Listener,
    modifier: Modifier = Modifier
) {
    @Composable
    fun RowScope.KeyboardButton(
        onClick: () -> Unit,
        content: String
    ) {
        Button(
            onClick = onClick,
            enabled = enabled,
            content = { Text(content, style = MaterialTheme.typography.displayMedium) },
            shape = ShapeDefaults.Small,
            modifier = Modifier
                .weight(1f)
                .fillMaxHeight()
                .padding(8.dp)
        )
    }

    @Composable
    fun RowScope.KeyboardIcon(imageVector: ImageVector, tint: Color) {
        Icon(
            imageVector,
            null,
            tint = tint,
            modifier = Modifier.weight(1f).fillMaxHeight()
        )
    }

    @Composable
    fun ColumnScope.KeyboardRow(content: @Composable RowScope.() -> Unit) {
        Row(
            content = content,
            modifier = Modifier.weight(1f)
        )
    }

    @Composable
    fun RowScope.KeyboardDigit(digit: Int) {
        KeyboardButton(
            onClick = { listener.onDigitEntered(digit) },
            content = digit.toString()
        )
    }

    @Composable
    fun RowScope.KeyboardDigit(row: Int, column: Int) {
        if (row < 0 || row > 3) throw IllegalArgumentException()
        if (column < 0 || column > 2) throw IllegalArgumentException()
        if (row == 3 && column != 0) throw IllegalArgumentException()

        val digit =
            if (row == 3) 0
            else when (configuration.type) {
                InputConfiguration.KeyboardType.Phone -> row * 3 + column + 1
                InputConfiguration.KeyboardType.Calculator -> (2 - row) * 3 + column + 1
            } % 10

        KeyboardDigit(digit)
    }

    val content = @Composable {
        Column {
            for (row in 0..2) {
                KeyboardRow {
                    for (column in 0..2) {
                        KeyboardDigit(row, column)
                    }
                }
            }

            KeyboardRow {
                val deleteButton = @Composable {
                    KeyboardButton(
                        onClick = { listener.onDeleteClicked() },
                        content = "<"
                    )
                }

                val finalDigit = @Composable {
                    KeyboardDigit(3, 0)
                }

                val confirmButton = @Composable {
                    when (confirmButton) {
                        ConfirmButton.Regular -> KeyboardButton(
                            onClick = { listener.onConfirmClicked() },
                            content = ">"
                        )

                        is ConfirmButton.Special -> KeyboardButton(
                            onClick = { listener.onConfirmClicked() },
                            content = confirmButton.label
                        )

                        ConfirmButton.ShowRight -> KeyboardIcon(
                            Icons.Default.Check,
                            Color(0, 255, 0),
                        )

                        ConfirmButton.ShowWrong -> KeyboardIcon(
                            Icons.Default.Close,
                            Color(255, 0, 0)
                        )
                    }
                }

                when (configuration.confirmButtonLocation) {
                    InputConfiguration.ConfirmButtonLocation.Left -> {
                        confirmButton()
                        finalDigit()
                        deleteButton()
                    }

                    InputConfiguration.ConfirmButtonLocation.Right -> {
                        deleteButton()
                        finalDigit()
                        confirmButton()
                    }
                }
            }
        }
    }

    val sizedContent = @Composable { modifier: Modifier ->
        Layout(
            content = content,
            modifier = modifier,
            measurePolicy = object: MeasurePolicy {
                override fun MeasureScope.measure(measurables: List<Measurable>, constraints: Constraints): MeasureResult {
                    val measurable = measurables.single()

                    val keyboardWidth = measurable.minIntrinsicWidth(Constraints.Infinity)
                    val keyboardHeight = measurable.minIntrinsicHeight(Constraints.Infinity)

                    val maxWidth = constraints.maxWidth
                    val maxHeight = constraints.maxHeight

                    val maxWidthFactor = maxWidth * 100 / keyboardWidth
                    val maxHeightFactor = maxHeight * 100 / keyboardHeight
                    val maxFactor = maxWidthFactor.coerceAtMost(maxHeightFactor)

                    val usedScaledFactor = if (maxFactor > 100) 100 + configuration.size * (maxFactor - 100) / 100 else 100

                    val placeable = measurable.measure(Constraints.fixed(
                        keyboardWidth * usedScaledFactor / 100,
                        keyboardHeight * usedScaledFactor / 100
                    ))

                    return layout(keyboardWidth * usedScaledFactor / 100, keyboardHeight * usedScaledFactor / 100) {
                        placeable.placeRelative(0, 0)
                    }
                }
            }
        )
    }

    Box(modifier) {
        sizedContent(
            Modifier.align(BiasAlignment(
                when (configuration.horizontal) {
                    InputConfiguration.HorizontalAlignment.Left -> -1f
                    InputConfiguration.HorizontalAlignment.Center -> 0f
                    InputConfiguration.HorizontalAlignment.Right -> 1f
                },
                when (configuration.vertical) {
                    InputConfiguration.VerticalAlignment.Top -> -1f
                    InputConfiguration.VerticalAlignment.Center -> 0f
                    InputConfiguration.VerticalAlignment.Bottom -> 1f
                }
            ))
        )
    }
}

interface Listener {
    fun onDigitEntered(digit: Int)
    fun onDeleteClicked()
    fun onConfirmClicked()
}

sealed class ConfirmButton {
    object Regular: ConfirmButton()
    data class Special(val label: String): ConfirmButton()
    object ShowRight: ConfirmButton()
    object ShowWrong: ConfirmButton()
}