/*
 * Calculate Android Copyright <C> 2020 - 2025 Jonas Lochmann
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation version 3 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see https://www.gnu.org/licenses/
 */
package com.rechnen.app.ui.training

import android.annotation.SuppressLint
import androidx.compose.foundation.layout.BoxWithConstraints
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.Row
import androidx.compose.foundation.layout.fillMaxWidth
import androidx.compose.foundation.layout.height
import androidx.compose.material3.MaterialTheme
import androidx.compose.material3.Text
import androidx.compose.runtime.Composable
import androidx.compose.runtime.remember
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.platform.LocalDensity
import androidx.compose.ui.text.rememberTextMeasurer
import androidx.compose.ui.text.style.TextAlign
import androidx.compose.ui.text.style.TextOverflow
import androidx.compose.ui.unit.dp
import androidx.compose.ui.unit.sp

data class TaskViewLine(val left: String, val right: String, val ellipsize: Boolean = false)
data class TaskViewLines(val lines: List<TaskViewLine>)

@SuppressLint("UnusedBoxWithConstraintsScope")
@Composable
fun TaskView(lines: TaskViewLines, modifier: Modifier = Modifier) {
    with(LocalDensity.current) {
        val maxFontSize = 50.sp
        val height = with(LocalDensity.current) { maxFontSize.toDp() * lines.lines.size }

        val textMeasurer = rememberTextMeasurer()
        val baseTextStyle = MaterialTheme.typography.labelMedium.copy(fontSize = maxFontSize)

        val longestLineWidth = remember(lines, textMeasurer, baseTextStyle) {
            (lines.lines.maxOfOrNull {
                if (it.ellipsize) 0
                else arrayOf(it.left, it.right).sumOf {
                    textMeasurer.measure(it, baseTextStyle).size.width
                }
            } ?: 0).toDp().coerceAtLeast(1.dp)
        }

        BoxWithConstraints(
            modifier
                .fillMaxWidth()
                .height(height),
            content = {
                val usableWidth = maxWidth

                val fontSizeFactor = (usableWidth / longestLineWidth).coerceAtMost(1f)
                val usedFontSize = (maxFontSize * fontSizeFactor)
                val usedTextStyle = baseTextStyle.copy(fontSize = usedFontSize)

                Column {
                    for (line in lines.lines) {
                        Row (
                            Modifier.height(maxFontSize.toDp()),
                            verticalAlignment = Alignment.CenterVertically
                        ) {
                            Text(
                                line.left,
                                style = usedTextStyle,
                                textAlign = TextAlign.Start,
                                maxLines = 1
                            )

                            Text(
                                line.right,
                                style = usedTextStyle,
                                textAlign = TextAlign.End,
                                maxLines = 1,
                                overflow = TextOverflow.Ellipsis,
                                modifier = Modifier.fillMaxWidth()
                            )
                        }
                    }
                }
            }
        )
    }
}