/*
 * Calculate Android Copyright <C> 2020 - 2025 Jonas Lochmann
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation version 3 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see https://www.gnu.org/licenses/
 */
package com.rechnen.app.ui.training

import android.content.Context
import android.content.Intent
import android.os.Bundle
import androidx.activity.compose.setContent
import androidx.activity.enableEdgeToEdge
import androidx.activity.viewModels
import androidx.compose.animation.core.tween
import androidx.compose.foundation.gestures.animateScrollBy
import androidx.compose.foundation.layout.Box
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.Row
import androidx.compose.foundation.layout.Spacer
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.foundation.layout.fillMaxWidth
import androidx.compose.foundation.layout.height
import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.layout.size
import androidx.compose.foundation.lazy.LazyColumn
import androidx.compose.foundation.lazy.items
import androidx.compose.foundation.lazy.rememberLazyListState
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.automirrored.filled.TrendingDown
import androidx.compose.material.icons.automirrored.filled.TrendingFlat
import androidx.compose.material.icons.automirrored.filled.TrendingUp
import androidx.compose.material.icons.filled.Check
import androidx.compose.material.icons.filled.Close
import androidx.compose.material3.Button
import androidx.compose.material3.Icon
import androidx.compose.material3.LinearProgressIndicator
import androidx.compose.material3.MaterialTheme
import androidx.compose.material3.Scaffold
import androidx.compose.material3.Text
import androidx.compose.runtime.Composable
import androidx.compose.runtime.LaunchedEffect
import androidx.compose.runtime.getValue
import androidx.compose.runtime.livedata.observeAsState
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.graphics.Color
import androidx.compose.ui.platform.LocalDensity
import androidx.compose.ui.res.stringResource
import androidx.compose.ui.text.style.TextAlign
import androidx.compose.ui.unit.dp
import androidx.fragment.app.FragmentActivity

import com.rechnen.app.R
import com.rechnen.app.data.AppDatabase
import com.rechnen.app.data.modelparts.InputConfiguration
import com.rechnen.app.ui.AppTheme
import com.rechnen.app.ui.tasktype.TaskTypeInfo

class TrainingMainActivity : FragmentActivity() {
    companion object {
        private const val EXTRA_USER_ID = "userId"
        private const val EXTRA_MODE = "mode"

        fun launch(context: Context, userId: Int, mode: TrainingMode) {
            context.startActivity(
                    Intent(context, TrainingMainActivity::class.java)
                            .putExtra(EXTRA_USER_ID, userId)
                            .putExtra(EXTRA_MODE, mode)
            )
        }

        @Composable
        fun TrainingScreen(
            progress: () -> Float,
            lines: TaskViewLines,
            keyboardConfiguration: InputConfiguration,
            keyboardConfirmButton: ConfirmButton,
            keyboardEnabled: Boolean,
            keyboardListener: Listener,
            modifier: Modifier = Modifier
        ) {
            Column (modifier) {
                LinearProgressIndicator(
                    progress = progress,
                    modifier = Modifier.padding(8.dp).fillMaxWidth()
                )

                TaskView(lines, Modifier.padding(16.dp))

                KeyboardView(
                    configuration = keyboardConfiguration,
                    confirmButton = keyboardConfirmButton,
                    enabled = keyboardEnabled,
                    listener = keyboardListener,
                    modifier = Modifier.weight(1f).fillMaxWidth().padding(8.dp)
                )
            }
        }
    }

    private val model: TrainingModel by viewModels()

    private lateinit var playtimeSaver: PlaytimeSaver

    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)

        val database = AppDatabase.with(this)
        val userId = intent.extras!!.getInt(EXTRA_USER_ID)
        val mode = intent.extras!!.getSerializable(EXTRA_MODE) as TrainingMode

        playtimeSaver = PlaytimeSaver(userId, database)

        model.init(
                database = database,
                userId = userId,
                mode = mode
        )

        enableEdgeToEdge()

        setContent {
            val trainingState by model.trainingState.observeAsState()
            val inputState by model.inputState.observeAsState()

            AppTheme {
                Scaffold { insets ->
                    when (val trainingState = trainingState) {
                        is TrainingState.ShowTask -> {
                            TrainingScreen(
                                progress = {
                                    when (trainingState) {
                                        is TrainingState.ShowTask.RegularModeTask -> {
                                            (trainingState.currentQuestionIndex + 1).toFloat() / trainingState.totalQuestionCounter
                                        }
                                        is TrainingState.ShowTask.TimeTrailTask -> {
                                            trainingState.elapsedTime.toFloat() / trainingState.totalTime
                                        }
                                    }
                                },
                                lines = trainingState.task.taskPresentation.let { presentation ->
                                    val resultLine = TaskViewLine(left = "=", right = inputState?.toString() ?: "", ellipsize = true)
                                    val emptyLine = TaskViewLine(left = "", right = "")

                                    if (presentation.size == 1)
                                        TaskViewLines(presentation + listOf(resultLine, emptyLine))
                                    else
                                        TaskViewLines(presentation + listOf(resultLine))
                                },
                                keyboardConfiguration = trainingState.inputConfiguration,
                                keyboardConfirmButton = when (trainingState.state) {
                                    TrainingShowTaskMode.FeedbackRight -> ConfirmButton.ShowRight
                                    TrainingShowTaskMode.FeedbackWrong -> ConfirmButton.ShowWrong
                                    TrainingShowTaskMode.WaitingForInput -> when (inputState?.specialInputType) {
                                        null -> ConfirmButton.Regular
                                        SpecialInputType.Nothing -> ConfirmButton.Regular
                                        SpecialInputType.NextItem -> ConfirmButton.Special(trainingState.task.separator!!)
                                        SpecialInputType.TogglePrefix -> ConfirmButton.Special(trainingState.task.prefix!!)
                                    }
                                },
                                keyboardEnabled = trainingState.state == TrainingShowTaskMode.WaitingForInput,
                                keyboardListener = object: Listener {
                                    override fun onDigitEntered(digit: Int) { playtimeSaver.startOrContinue(); model.handleNumberButton(digit) }
                                    override fun onConfirmClicked() { playtimeSaver.startOrContinue(); model.handleConfirmButton() }
                                    override fun onDeleteClicked() { playtimeSaver.startOrContinue(); model.handleDeleteButton() }
                                },
                                modifier = Modifier.padding(insets)
                            )
                        }
                        is TrainingState.EndOfRound -> {
                            Column (Modifier.padding(insets)) {
                                val listState = rememberLazyListState()
                                val startOffset = with(LocalDensity.current) { 128.dp.toPx().toInt() }

                                LaunchedEffect(trainingState, startOffset) {
                                    listState.scrollToItem(0, startOffset)
                                    listState.animateScrollBy(-startOffset.toFloat(), tween(1000))
                                }

                                LazyColumn (Modifier.weight(1f), state = listState) {
                                    item {
                                        val content = when (trainingState) {
                                            is TrainingState.EndOfRound.RegularEndOfRound -> when (trainingState.result) {
                                                TrainingRoundEndState.EverythingGood -> if (trainingState.mistakes == 0)
                                                    stringResource(R.string.training_block_result_perfect)
                                                else
                                                    resources.getQuantityString(R.plurals.training_block_result_good, trainingState.mistakes, trainingState.mistakes)
                                                TrainingRoundEndState.TooMuchMistakes -> resources.getQuantityString(R.plurals.training_block_result_too_much_mistakes, trainingState.mistakes, trainingState.mistakes)
                                                TrainingRoundEndState.TooSlow -> stringResource(R.string.training_block_result_too_slow, formatMistakes(trainingState.mistakes), formatSeconds((trainingState.time / 1000).toInt()))
                                            }
                                            is TrainingState.EndOfRound.TimeTrailEndOfRound -> {
                                                val totalTasks = trainingState.tasks.size
                                                val correctTasks = trainingState.tasks.filter { it.correct }.size

                                                stringResource(
                                                    R.string.time_trial_result_headline,
                                                    resources.getQuantityString(R.plurals.difficulty_general_summary_tasks, totalTasks, totalTasks),
                                                    resources.getQuantityString(R.plurals.difficulty_general_summary_tasks, correctTasks, correctTasks)
                                                )
                                            }
                                        }

                                        Text(
                                            content,
                                            Modifier.fillMaxWidth().padding(16.dp),
                                            textAlign = TextAlign.Center,
                                            style = MaterialTheme.typography.headlineMedium
                                        )
                                    }

                                    when (trainingState) {
                                        is TrainingState.EndOfRound.RegularEndOfRound -> {
                                            val maxAvgDuration =
                                                trainingState.statistic.filter { it.parameter == TrainingStatisticItemParameter.AverageDurationPerCorrectTask }
                                                    .maxOfOrNull { inputRow ->
                                                        inputRow.values.values.maxOrNull() ?: 0
                                                    } ?: 0

                                            items(trainingState.statistic) {
                                                Column {
                                                    Row(verticalAlignment = Alignment.CenterVertically) {
                                                        Text(
                                                            stringResource(
                                                                when (it.parameter) {
                                                                    TrainingStatisticItemParameter.AverageDurationPerCorrectTask -> R.string.statistic_average_duration_per_correct_task_at
                                                                    TrainingStatisticItemParameter.PercentageOfCorrectlySolvedTasks ->  R.string.statistic_percentage_of_correctly_solved_tasks
                                                                },
                                                                stringResource(TaskTypeInfo.get(it.taskType).labelResourceId)
                                                            ),
                                                            Modifier.padding(8.dp).weight(1f),
                                                            textAlign = TextAlign.Center,
                                                            style = MaterialTheme.typography.titleMedium
                                                        )

                                                        if (it.change != null) Icon(
                                                            when (it.change) {
                                                                TrainingStatisticItemChange.Better -> Icons.AutoMirrored.Filled.TrendingUp
                                                                TrainingStatisticItemChange.Equal -> Icons.AutoMirrored.Filled.TrendingFlat
                                                                TrainingStatisticItemChange.Worser -> Icons.AutoMirrored.Filled.TrendingDown
                                                            },
                                                            null,
                                                            Modifier.size(24.dp).padding(8.dp)
                                                        )
                                                    }

                                                    for (item in it.values.entries.sortedBy { TrainingStatisticItemSourceSorting.rank(it.key) }) {
                                                        val sourceLabel = stringResource(
                                                            when (item.key) {
                                                                TrainingStatisticItemSource.BeforeThisSession -> R.string.statistic_source_before_this_session
                                                                TrainingStatisticItemSource.ThisSessionBeforeThisRound -> R.string.statistic_source_this_session_before_this_round
                                                                TrainingStatisticItemSource.ThisRound -> R.string.statistic_source_this_round
                                                            }
                                                        )

                                                        val formattedValue = when (it.parameter) {
                                                            TrainingStatisticItemParameter.AverageDurationPerCorrectTask -> formatMilliseconds(item.value.toLong())
                                                            TrainingStatisticItemParameter.PercentageOfCorrectlySolvedTasks -> "${item.value}%"
                                                        }

                                                        Text("${sourceLabel}: $formattedValue", Modifier.padding(8.dp))

                                                        val max = when (it.parameter) {
                                                            TrainingStatisticItemParameter.PercentageOfCorrectlySolvedTasks -> 100
                                                            TrainingStatisticItemParameter.AverageDurationPerCorrectTask -> maxAvgDuration
                                                        }

                                                        LinearProgressIndicator(
                                                            progress = { item.value.toFloat() / max.toFloat() },
                                                            modifier = Modifier.fillMaxWidth().padding(8.dp)
                                                        )
                                                    }
                                                }
                                            }
                                        }
                                        is TrainingState.EndOfRound.TimeTrailEndOfRound -> {
                                            items(trainingState.tasks) { item ->
                                                Row (
                                                    Modifier.padding(8.dp),
                                                    verticalAlignment = Alignment.CenterVertically
                                                ) {
                                                    Column (Modifier.weight(1f)) {
                                                        Text(
                                                            "${item.task.readableTask}=${item.input.format(item.task)}"
                                                        )

                                                        Text(
                                                            formatMilliseconds(item.time),
                                                            style = MaterialTheme.typography.bodyMedium
                                                        )
                                                    }

                                                    if (item.correct) Icon(
                                                        Icons.Default.Check,
                                                        null,
                                                        tint = Color(0, 255, 0),
                                                        modifier = Modifier.padding(8.dp)
                                                    ) else Icon(
                                                        Icons.Default.Close,
                                                        null,
                                                        tint = Color(255, 0, 0),
                                                        modifier = Modifier.padding(8.dp)
                                                    )
                                                }
                                            }
                                        }
                                    }.let {/* require handling all paths */}

                                    if (trainingState is TrainingState.EndOfRound.RegularEndOfRound) {
                                        item {
                                            Text(
                                                stringResource(R.string.training_result_difficulty_tip),
                                                Modifier.fillMaxWidth().padding(16.dp),
                                                textAlign = TextAlign.Center
                                            )
                                        }
                                    }
                                }

                                val wouldContinueWithNewTasks = when (trainingState) {
                                    is TrainingState.EndOfRound.RegularEndOfRound -> trainingState.result == TrainingRoundEndState.EverythingGood
                                    is TrainingState.EndOfRound.TimeTrailEndOfRound -> true
                                }

                                Button(
                                    onClick = {
                                        playtimeSaver.startOrContinue()

                                        model.handleContinueButton()
                                    },
                                    content = {
                                        if (wouldContinueWithNewTasks) Text(stringResource(R.string.training_block_result_new_tasks))
                                        else Text(stringResource(R.string.training_block_result_redo))
                                    },
                                    modifier = Modifier.padding(horizontal = 16.dp, vertical = 8.dp).align(Alignment.End)
                                )
                            }
                        }
                        is TrainingState.Countdown -> {
                            Box(Modifier.fillMaxSize().padding(insets)) {
                                Column (Modifier.fillMaxWidth().align(Alignment.CenterStart)) {
                                    Text(
                                        trainingState.remainingSeconds.toString(),
                                        style = MaterialTheme.typography.headlineMedium,
                                        textAlign = TextAlign.Center,
                                        modifier = Modifier.fillMaxWidth()
                                    )

                                    Spacer(Modifier.height(16.dp))

                                    Text(
                                        trainingState.info ?: "",
                                        textAlign = TextAlign.Center,
                                        modifier = Modifier.fillMaxWidth()
                                    )
                                }
                            }
                        }
                        null -> {/* show nothing */}
                    }.let {/* require handling all cases */}
                }
            }
        }

        model.trainingState.observe(this) { state ->
            when (state) {
                is TrainingState.ShowTask -> playtimeSaver.startOrContinue()
                is TrainingState.EndOfRound -> playtimeSaver.stop()
                is TrainingState.Countdown -> playtimeSaver.stop()
            }.let {/* require handling all cases */}
        }
    }

    override fun onStop() {
        super.onStop()

        playtimeSaver.stop()
    }

    private fun formatMistakes(mistakes: Int) = resources.getQuantityString(R.plurals.training_result_mistakes, mistakes, mistakes)
    private fun formatSeconds(seconds: Int) = resources.getQuantityString(R.plurals.training_result_seconds, seconds, seconds)
    private fun formatMilliseconds(milliseconds: Long) = "${milliseconds / 1000}.${milliseconds / 100 % 10} ${getString(R.string.statistic_value_seconds)}"
}
