/*
 * Calculate Android Copyright <C> 2020 - 2025 Jonas Lochmann
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation version 3 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see https://www.gnu.org/licenses/
 */
package com.rechnen.app.ui.user

import androidx.compose.foundation.clickable
import androidx.compose.foundation.layout.Arrangement
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.Row
import androidx.compose.foundation.layout.defaultMinSize
import androidx.compose.foundation.layout.fillMaxWidth
import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.layout.width
import androidx.compose.foundation.text.KeyboardOptions
import androidx.compose.material3.ExperimentalMaterial3Api
import androidx.compose.material3.ModalBottomSheet
import androidx.compose.material3.OutlinedTextField
import androidx.compose.material3.Slider
import androidx.compose.material3.Switch
import androidx.compose.material3.Text
import androidx.compose.runtime.Composable
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.runtime.setValue
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.platform.LocalContext
import androidx.compose.ui.res.stringResource
import androidx.compose.ui.text.input.ImeAction
import androidx.compose.ui.text.input.KeyboardType
import androidx.compose.ui.unit.dp
import com.rechnen.app.R
import com.rechnen.app.data.modelparts.*
import com.rechnen.app.extension.seekBarIndexOf
import kotlin.math.roundToInt

object EditUserDifficultyDialog {
    @OptIn(ExperimentalMaterial3Api::class)
    @Composable
    fun Main(
        difficulty: TaskDifficulty,
        listener: Listener
    ) {
        val withTenTransgression = stringResource(R.string.suffix_with_ten_transgression)
        val withoutTenTransgression = stringResource(R.string.suffix_without_ten_transgression)

        val difficultyProfile = TaskDifficultyProfiles.profilesAndLabels.find { it.first == difficulty }

        var subscreen by rememberSaveable { mutableStateOf(null as Subscreen?) }

        ModalBottomSheet(
            onDismissRequest = { listener.dismiss() },
            content = {
                Column {
                    BottomSheet.ListItem(
                        title = stringResource(R.string.difficulty_profile),
                        subtitle = stringResource(difficultyProfile?.second ?: R.string.difficulty_profile_custom),
                        listener = { subscreen = Subscreen.Profile }
                    )

                    BottomSheet.ListItem(
                        title = stringResource(R.string.difficulty_operation_addition),
                        subtitle = if (difficulty.addition.enable) {
                            stringResource(
                                R.string.difficulty_operation_addition_enabled,
                                formatDigitNumber(difficulty.addition.digitsOfFirstNumber),
                                formatDigitNumber(difficulty.addition.digitsOfSecondNumber),
                                if (difficulty.addition.digitsOfSecondNumber > 1)
                                    ""
                                else if (difficulty.addition.enableTenTransgression)
                                    withTenTransgression
                                else
                                    withoutTenTransgression
                            )
                        } else {
                            stringResource(R.string.difficulty_operation_disabled)
                        },
                        listener = { subscreen = Subscreen.Addition }
                    )

                    BottomSheet.ListItem(
                        title = stringResource(R.string.difficulty_operation_subtraction),
                        subtitle = if (difficulty.subtraction.enable) {
                            stringResource(
                                R.string.difficulty_operation_subtraction_enabled,
                                formatDigitNumber(difficulty.subtraction.digitsOfFirstNumber),
                                formatDigitNumber(difficulty.subtraction.digitsOfSecondNumber),
                                if (difficulty.subtraction.allowNegativeResults)
                                    stringResource(R.string.suffix_with_negative_results)
                                else if (difficulty.subtraction.digitsOfSecondNumber > 1)
                                    ""
                                else if (difficulty.subtraction.enableTenTransgression)
                                    withTenTransgression
                                else
                                    withoutTenTransgression
                            )
                        } else {
                            stringResource(R.string.difficulty_operation_disabled)
                        },
                        listener = { subscreen = Subscreen.Subtraction }
                    )

                    BottomSheet.ListItem(
                        title = stringResource(R.string.difficulty_operation_multiplication),
                        subtitle = if (difficulty.multiplication.enable) {
                            stringResource(
                                R.string.difficulty_operation_multiplication_enabled,
                                formatDigitNumber(difficulty.multiplication.digitsOfFirstNumber),
                                formatDigitNumber(difficulty.multiplication.digitsOfSecondNumber)
                            )
                        } else {
                            stringResource(R.string.difficulty_operation_disabled)
                        },
                        listener = { subscreen = Subscreen.Multiplication }
                    )

                    BottomSheet.ListItem(
                        title = stringResource(R.string.difficulty_operation_division),
                        subtitle = if (difficulty.division.enable) {
                            stringResource(
                                R.string.difficulty_operation_division_enabled,
                                difficulty.division.maxDividend,
                                difficulty.division.maxDivisor,
                                stringResource(if (difficulty.division.withRemainder) R.string.suffix_with_remainder else R.string.suffix_without_remainder)
                            )
                        } else {
                            stringResource(R.string.difficulty_operation_disabled)
                        },
                        listener = { subscreen = Subscreen.Division }
                    )

                    BottomSheet.ListItem(
                        title = stringResource(R.string.difficulty_operation_factorization),
                        subtitle = if (difficulty.factorization.enable) {
                            stringResource(R.string.difficulty_operation_factorization_enabled, difficulty.factorization.smallestNumber, difficulty.factorization.biggestNumber)
                        } else {
                            stringResource(R.string.difficulty_operation_disabled)
                        },
                        listener = { subscreen = Subscreen.PrimeFactorization }
                    )

                    BottomSheet.ListItem(
                        title = stringResource(R.string.difficulty_general_settings),
                        subtitle = stringResource(
                            R.string.difficulty_general_summary,
                            LocalContext.current.resources.getQuantityString(R.plurals.difficulty_general_summary_tasks, difficulty.roundConfig.tasksPerRound, difficulty.roundConfig.tasksPerRound),
                            LocalContext.current.resources.getQuantityString(
                                R.plurals.training_result_seconds,
                                difficulty.roundConfig.maximalTimePerRound / 1000,
                                difficulty.roundConfig.maximalTimePerRound / 1000
                            ),
                            LocalContext.current.resources.getQuantityString(R.plurals.difficulty_general_summary_mistake, difficulty.roundConfig.maximalMistakesPerRound, difficulty.roundConfig.maximalMistakesPerRound),
                            difficulty.roundConfig.reAskWrongInNextRound,
                            difficulty.roundConfig.reAskRightInNextRound
                        ),
                        listener = { subscreen = Subscreen.General }
                    )

                    BottomSheet.ListItem(
                        title = stringResource(R.string.generic_save),
                        subtitle = null,
                        listener = { listener.save() }
                    )
                }
            }
        )

        when (subscreen) {
            null -> {/* nothing to do */}
            Subscreen.Profile -> ModalBottomSheet(
                onDismissRequest = { subscreen = null },
                content = {
                    for (profile in TaskDifficultyProfiles.profilesAndLabels) {
                        BottomSheet.RadioButton(
                            stringResource(profile.second),
                            difficultyProfile == profile
                        ) {
                            listener.adjustDifficulty { profile.first }

                            subscreen = null
                        }
                    }
                }
            )
            Subscreen.Addition -> ModalBottomSheet(
                onDismissRequest = { subscreen = null },
                content = {
                    BottomSheet.Title(stringResource(R.string.difficulty_operation_addition))

                    EnableOperatorView(
                        status = getOperatorStatus(difficulty, TaskType.Addition),
                        onCheckedChange = { enable ->
                            listener.adjustDifficulty { difficulty ->
                                difficulty.copy(addition = difficulty.addition.copy(enable = enable))
                            }
                        }
                    )

                    DifficultyDialogSwitch(
                        label = stringResource(R.string.difficulty_enable_ten_transgression),
                        checked = difficulty.addition.enableTenTransgression,
                        onCheckedChange = { enable ->
                            listener.adjustDifficulty { difficulty ->
                                difficulty.copy(
                                    addition = difficulty.addition.copy(
                                        enableTenTransgression = enable,
                                        digitsOfSecondNumber = if (enable) difficulty.addition.digitsOfSecondNumber else 1
                                    )
                                )
                            }
                        },
                        enabled = difficulty.addition.enable
                    )

                    TwoOperandSliders(
                        difficulty.addition.digitsOfFirstNumber,
                        difficulty.addition.digitsOfSecondNumber,
                        { digits ->
                            listener.adjustDifficulty { difficulty ->
                                difficulty.copy(
                                    addition = difficulty.addition.copy(
                                        digitsOfFirstNumber = digits
                                    )
                                )
                            }
                        },
                        { digits ->
                            listener.adjustDifficulty { difficulty ->
                                difficulty.copy(
                                    addition = difficulty.addition.copy(
                                        digitsOfSecondNumber = digits,
                                        enableTenTransgression = difficulty.addition.enableTenTransgression || digits != 1
                                    )
                                )
                            }
                        },
                        enabled = difficulty.addition.enable
                    )
                }
            )
            Subscreen.Subtraction -> ModalBottomSheet(
                onDismissRequest = { subscreen = null },
                content = {
                    BottomSheet.Title(stringResource(R.string.difficulty_operation_subtraction))

                    EnableOperatorView(
                        status = getOperatorStatus(difficulty, TaskType.Subtraction),
                        onCheckedChange = { enable ->
                            listener.adjustDifficulty { difficulty ->
                                difficulty.copy(subtraction = difficulty.subtraction.copy(enable = enable))
                            }
                        }
                    )

                    DifficultyDialogSwitch(
                        label = stringResource(R.string.difficulty_enable_ten_transgression),
                        checked = difficulty.subtraction.enableTenTransgression,
                        onCheckedChange = { enable ->
                            listener.adjustDifficulty { difficulty ->
                                difficulty.copy(
                                    subtraction = difficulty.subtraction.copy(
                                        enableTenTransgression = enable,
                                        digitsOfSecondNumber = if (enable) difficulty.subtraction.digitsOfSecondNumber else 1
                                    )
                                )
                            }
                        },
                        enabled = difficulty.subtraction.enable
                    )

                    DifficultyDialogSwitch(
                        label = stringResource(R.string.difficulty_allow_negative_results),
                        checked = difficulty.subtraction.allowNegativeResults,
                        onCheckedChange = { enable ->
                            listener.adjustDifficulty { difficulty ->
                                difficulty.copy(
                                    subtraction = difficulty.subtraction.copy(
                                        allowNegativeResults = enable,
                                        digitsOfSecondNumber = difficulty.subtraction.digitsOfSecondNumber.let {
                                            if (enable) it
                                            else it.coerceAtMost(difficulty.subtraction.digitsOfFirstNumber)
                                        }
                                    )
                                )
                            }
                        },
                        enabled = difficulty.subtraction.enable
                    )

                    TwoOperandSliders(
                        difficulty.subtraction.digitsOfFirstNumber,
                        difficulty.subtraction.digitsOfSecondNumber,
                        { digits ->
                            listener.adjustDifficulty { difficulty ->
                                difficulty.copy(
                                    subtraction = difficulty.subtraction.copy(
                                        digitsOfFirstNumber = digits,
                                        allowNegativeResults = difficulty.subtraction.allowNegativeResults || difficulty.subtraction.digitsOfFirstNumber < difficulty.subtraction.digitsOfSecondNumber
                                    )
                                )
                            }
                        },
                        { digits ->
                            listener.adjustDifficulty { difficulty ->
                                difficulty.copy(
                                    subtraction = difficulty.subtraction.copy(
                                        digitsOfSecondNumber = digits,
                                        enableTenTransgression = difficulty.subtraction.enableTenTransgression || digits != 1,
                                        allowNegativeResults = difficulty.subtraction.allowNegativeResults || digits > difficulty.subtraction.digitsOfFirstNumber
                                    )
                                )
                            }
                        },
                        enabled = difficulty.subtraction.enable
                    )
                }
            )
            Subscreen.Multiplication -> ModalBottomSheet(
                onDismissRequest = { subscreen = null },
                content = {
                    BottomSheet.Title(stringResource(R.string.difficulty_operation_multiplication))

                    EnableOperatorView(
                        status = getOperatorStatus(difficulty, TaskType.Multiplication),
                        onCheckedChange = { enable ->
                            listener.adjustDifficulty { difficulty ->
                                difficulty.copy(multiplication = difficulty.multiplication.copy(enable = enable))
                            }
                        }
                    )

                    TwoOperandSliders(
                        difficulty.multiplication.digitsOfFirstNumber,
                        difficulty.multiplication.digitsOfSecondNumber,
                        { digits ->
                            listener.adjustDifficulty { difficulty ->
                                difficulty.copy(
                                    multiplication = difficulty.multiplication.copy(
                                        digitsOfFirstNumber = digits
                                    )
                                )
                            }
                        },
                        { digits ->
                            listener.adjustDifficulty { difficulty ->
                                difficulty.copy(
                                    multiplication = difficulty.multiplication.copy(
                                        digitsOfSecondNumber = digits
                                    )
                                )
                            }
                        },
                        enabled = difficulty.multiplication.enable
                    )
                }
            )
            Subscreen.Division -> ModalBottomSheet(
                onDismissRequest = { subscreen = null },
                content = {
                    BottomSheet.Title(stringResource(R.string.difficulty_operation_division))

                    EnableOperatorView(
                        status = getOperatorStatus(difficulty, TaskType.Division),
                        onCheckedChange = { enable ->
                            listener.adjustDifficulty { difficulty ->
                                difficulty.copy(division = difficulty.division.copy(enable = enable))
                            }
                        }
                    )

                    DifficultyDialogSwitch(
                        label = stringResource(R.string.difficulty_allow_remainder),
                        checked = difficulty.division.withRemainder,
                        onCheckedChange = { enable ->
                            listener.adjustDifficulty { difficulty ->
                                difficulty.copy(
                                    division = difficulty.division.copy(
                                        withRemainder = enable
                                    )
                                )
                            }
                        },
                        enabled = difficulty.division.enable
                    )

                    SliderView(
                        { stringResource(R.string.difficulty_max_dividend, DivisionConfig.MAX_DIVIDEND_SUGGESTIONS[it]) },
                        DivisionConfig.MAX_DIVIDEND_SUGGESTIONS.seekBarIndexOf(difficulty.division.maxDividend),
                        0,
                        DivisionConfig.MAX_DIVIDEND_SUGGESTIONS.size - 1,
                        { index ->
                            val newDividend = DivisionConfig.MAX_DIVIDEND_SUGGESTIONS[index]

                            listener.adjustDifficulty { difficulty ->
                                difficulty.copy(
                                    division = difficulty.division.copy(
                                        maxDividend = newDividend,
                                        maxDivisor = difficulty.division.maxDivisor.coerceAtMost(newDividend)
                                    )
                                )
                            }
                        },
                        difficulty.division.enable
                    )

                    SliderView(
                        { stringResource(R.string.difficulty_max_divisor, DivisionConfig.MAX_DIVISOR_SUGGESTIONS[it]) },
                        DivisionConfig.MAX_DIVISOR_SUGGESTIONS.seekBarIndexOf(difficulty.division.maxDivisor),
                        0,
                        DivisionConfig.MAX_DIVISOR_SUGGESTIONS.size - 1,
                        { index ->
                            val newDivisor = DivisionConfig.MAX_DIVISOR_SUGGESTIONS[index]

                            listener.adjustDifficulty { difficulty ->
                                difficulty.copy(
                                    division = difficulty.division.copy(
                                        maxDivisor = newDivisor,
                                        maxDividend = difficulty.division.maxDividend.coerceAtLeast(newDivisor)
                                    )
                                )
                            }
                        },
                        difficulty.division.enable
                    )
                }
            )
            Subscreen.PrimeFactorization -> ModalBottomSheet(
                onDismissRequest = { subscreen = null },
                content = {
                    Column {
                        BottomSheet.Title(stringResource(R.string.difficulty_operation_factorization))

                        EnableOperatorView(
                            status = getOperatorStatus(difficulty, TaskType.Factorization),
                            onCheckedChange = { enable ->
                                listener.adjustDifficulty { difficulty ->
                                    difficulty.copy(factorization = difficulty.factorization.copy(enable = enable))
                                }
                            }
                        )

                        NumberRow(
                            label = stringResource(R.string.difficulty_operation_min_number),
                            value = difficulty.factorization.smallestNumber,
                            enabled = difficulty.factorization.enable,
                            onUpdateValue = { newValue ->
                                if (newValue < FactorizationConfig.MIN_NUMBER || newValue > FactorizationConfig.MAX_NUMBER) false
                                else {
                                    listener.adjustDifficulty { difficulty ->
                                        difficulty.copy(
                                            factorization = difficulty.factorization.copy(
                                                smallestNumber = newValue,
                                                biggestNumber = difficulty.factorization.biggestNumber.coerceAtLeast(newValue)
                                            )
                                        )
                                    }

                                    true
                                }
                            }
                        )

                        NumberRow(
                            label = stringResource(R.string.difficulty_operation_max_number),
                            value = difficulty.factorization.biggestNumber,
                            enabled = difficulty.factorization.enable,
                            onUpdateValue = { newValue ->
                                if (newValue < FactorizationConfig.MIN_NUMBER || newValue > FactorizationConfig.MAX_NUMBER) false
                                else {
                                    listener.adjustDifficulty { difficulty ->
                                        difficulty.copy(
                                            factorization = difficulty.factorization.copy(
                                                biggestNumber = newValue,
                                                smallestNumber = difficulty.factorization.smallestNumber.coerceAtMost(newValue)
                                            )
                                        )
                                    }

                                    true
                                }
                            }
                        )
                    }
                }
            )
            Subscreen.General -> ModalBottomSheet(
                onDismissRequest = { subscreen = null },
                content = {
                    SliderView(
                        label = {
                            LocalContext.current.resources.getQuantityString(
                                R.plurals.difficulty_tasks_per_block,
                                it,
                                it
                            )
                        },
                        difficulty.roundConfig.tasksPerRound,
                        min = 5,
                        max = 50,
                        stepSize = 5,
                        onValueChanged = {
                            listener.adjustDifficulty { difficulty ->
                                difficulty.copy(
                                    roundConfig = difficulty.roundConfig.copy(
                                        tasksPerRound = it,
                                        maximalTimePerRound = difficulty.roundConfig.maximalTimePerRound * it / difficulty.roundConfig.tasksPerRound,
                                        maximalMistakesPerRound = difficulty.roundConfig.maximalMistakesPerRound.coerceAtMost(it),
                                        reAskRightInNextRound = difficulty.roundConfig.reAskRightInNextRound.coerceAtMost(it),
                                        reAskWrongInNextRound = difficulty.roundConfig.reAskWrongInNextRound.coerceAtMost(it)
                                    )
                                )
                            }
                        }
                    )

                    SliderView(
                        label = { stringResource(
                            R.string.difficulty_time_per_round,
                            LocalContext.current.resources.getQuantityString(R.plurals.training_result_seconds, it / 1000, it / 1000)
                        ) },
                        value = difficulty.roundConfig.maximalTimePerRound,
                        min = difficulty.roundConfig.tasksPerRound * 1000,
                        max = difficulty.roundConfig.tasksPerRound * 1000 * 10,
                        stepSize = difficulty.roundConfig.tasksPerRound * 1000,
                        onValueChanged = {
                            listener.adjustDifficulty { difficulty ->
                                difficulty.copy(
                                    roundConfig = difficulty.roundConfig.copy(
                                        maximalTimePerRound = it
                                    )
                                )
                            }
                        }
                    )

                    SliderView(
                        label = {
                            LocalContext.current.resources.getQuantityString(
                                R.plurals.difficulty_max_wrong_tasks,
                                it,
                                it
                            )
                        },
                        value = difficulty.roundConfig.maximalMistakesPerRound,
                        min = 0,
                        max = difficulty.roundConfig.tasksPerRound,
                        onValueChanged = {
                            listener.adjustDifficulty { difficulty ->
                                difficulty.copy(
                                    roundConfig = difficulty.roundConfig.copy(
                                        maximalMistakesPerRound = it,
                                        reAskWrongInNextRound = difficulty.roundConfig.reAskWrongInNextRound.coerceAtMost(it)
                                    )
                                )
                            }
                        }
                    )

                    SliderView(
                        label = {
                            LocalContext.current.resources.getQuantityString(
                                R.plurals.difficulty_take_right_tasks,
                                it,
                                it
                            )
                        },
                        value = difficulty.roundConfig.reAskRightInNextRound,
                        min = 0,
                        max = difficulty.roundConfig.tasksPerRound,
                        onValueChanged = {
                            listener.adjustDifficulty { difficulty ->
                                difficulty.copy(
                                    roundConfig = difficulty.roundConfig.copy(
                                        reAskRightInNextRound = it
                                    )
                                )
                            }
                        }
                    )

                    SliderView(
                        label = {
                            LocalContext.current.resources.getQuantityString(
                                R.plurals.difficulty_take_wrong_tasks,
                                it,
                                it
                            )
                        },
                        value = difficulty.roundConfig.reAskWrongInNextRound,
                        min = 0,
                        max = difficulty.roundConfig.maximalMistakesPerRound,
                        onValueChanged = {
                            listener.adjustDifficulty { difficulty ->
                                difficulty.copy(
                                    roundConfig = difficulty.roundConfig.copy(
                                        reAskWrongInNextRound = it
                                    )
                                )
                            }
                        }
                    )
                }
            )
        }
    }

    interface Listener {
        fun adjustDifficulty(updater: (TaskDifficulty) -> TaskDifficulty)
        fun save()
        fun dismiss()
    }

    private enum class Subscreen {
        Profile,
        Addition,
        Subtraction,
        Multiplication,
        Division,
        PrimeFactorization,
        General,
    }

    private enum class EnableOperatorStatus {
        Disabled, Enabled, LastEnabled
    }

    @Composable
    private fun DifficultyDialogSwitch(
        label: String,
        checked: Boolean,
        onCheckedChange: (Boolean) -> Unit,
        enabled: Boolean = true
    ) {
        Row(
            verticalAlignment = Alignment.CenterVertically,
            horizontalArrangement = Arrangement.spacedBy(16.dp),
            modifier = Modifier
                .fillMaxWidth()
                .defaultMinSize(minHeight = 32.dp)
                .clickable(enabled = enabled, onClick = { onCheckedChange(!checked) })
                .padding(vertical = 8.dp, horizontal = 16.dp)
        ) {
            Text(label, Modifier.weight(1f))

            Switch(
                checked = checked,
                onCheckedChange = onCheckedChange,
                enabled = enabled
            )
        }
    }

    @Composable
    private fun EnableOperatorView(status: EnableOperatorStatus, onCheckedChange: (Boolean) -> Unit) {
        DifficultyDialogSwitch(
            label = stringResource(R.string.difficulty_operation_enable_switch),
            checked = status != EnableOperatorStatus.Disabled,
            enabled = status != EnableOperatorStatus.LastEnabled,
            onCheckedChange = onCheckedChange
        )

        if (status == EnableOperatorStatus.LastEnabled) Text(
            stringResource(R.string.difficulty_operation_cannot_disable_reason),
            Modifier.padding(vertical = 8.dp, horizontal = 16.dp)
        )
    }

    @Composable
    private fun TwoOperandSliders(
        value1: Int,
        value2: Int,
        onFirstValueChanged: (Int) -> Unit,
        onSecondValueChanged: (Int) -> Unit,
        enabled: Boolean = true
    ) {
        SliderView({ stringResource(R.string.difficulty_operation_first_number_digits, formatDigitNumber(it)) }, value1, 1, 5, onFirstValueChanged, enabled)
        SliderView({ stringResource(R.string.difficulty_operation_second_number_digits, formatDigitNumber(it)) }, value2, 1, 5, onSecondValueChanged, enabled)
    }

    @Composable
    private fun SliderView(label: @Composable (Int) -> String, value: Int, min: Int, max: Int, onValueChanged: (Int) -> Unit, enabled: Boolean = true, stepSize: Int = 1) {
        Text(
            label(value),
            Modifier.padding(horizontal = 16.dp)
        )

        val steps = (max - min) / stepSize - 1

        Slider(
            value = value.toFloat(),
            onValueChange = { onValueChanged(it.roundToInt()) },
            steps = steps.coerceAtLeast(0),
            valueRange = min.toFloat()..max.toFloat(),
            enabled = enabled && max > min,
            modifier = Modifier
                .padding(horizontal = 32.dp)
                .padding(bottom = 4.dp)
        )
    }

    @Composable
    private fun NumberRow(label: String, value: Int, enabled: Boolean, onUpdateValue: (Int) -> Boolean) {
        Row (
            horizontalArrangement = Arrangement.SpaceBetween,
            verticalAlignment = Alignment.CenterVertically,
            modifier = Modifier
                .fillMaxWidth()
                .padding(horizontal = 16.dp)
                .padding(top = 8.dp)
        ) {
            Text(label, Modifier.padding(end = 16.dp))

            val src = value
            var value by rememberSaveable { mutableStateOf(Triple(-1, "", false)) }

            val (actualValue, actualError) = run {
                val (savedSrc, savedCur, savedValid) = value

                if (savedSrc == src) Pair(savedCur, savedValid)
                else Pair(src.toString(), false)
            }

            OutlinedTextField(
                value = actualValue,
                isError = actualError,
                singleLine = true,
                keyboardOptions = KeyboardOptions.Default.copy(
                    imeAction = ImeAction.Done,
                    keyboardType = KeyboardType.Number
                ),
                onValueChange = { newValue ->
                    val valueInt = newValue.toIntOrNull()

                    val isValid = if (valueInt == null) false else onUpdateValue(valueInt)

                    if (isValid) value = Triple(valueInt!!, newValue, false)
                    else value = Triple(src, newValue, true)
                },
                enabled = enabled,
                modifier = Modifier.width(128.dp)
            )
        }
    }

    @Composable
    private fun formatDigitNumber(n: Int) = LocalContext.current.resources.getQuantityString(R.plurals.difficulty_digit_counter, n, n)

    private fun getOperatorStatus(difficulty: TaskDifficulty, type: TaskType) = if (difficulty.enabledTaskTypes.contains(type))
        if (difficulty.enabledTaskTypes.size == 1) EnableOperatorStatus.LastEnabled else EnableOperatorStatus.Enabled
    else
        EnableOperatorStatus.Disabled
}