/*
 * Calculate Android Copyright <C> 2020 - 2025 Jonas Lochmann
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation version 3 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see https://www.gnu.org/licenses/
 */
package com.rechnen.app.ui.user

import androidx.compose.material3.ExperimentalMaterial3Api
import androidx.compose.material3.ModalBottomSheet
import androidx.compose.runtime.Composable
import androidx.compose.runtime.getValue
import androidx.compose.runtime.mutableStateOf
import androidx.compose.runtime.saveable.rememberSaveable
import androidx.compose.runtime.setValue
import androidx.compose.ui.res.stringResource
import com.rechnen.app.R
import com.rechnen.app.data.modelparts.InputConfiguration

object EditUserInputConfigurationDialog {
    @OptIn(ExperimentalMaterial3Api::class)
    @Composable
    fun Main(config: InputConfiguration, listener: Listener) {
        var subcreen by rememberSaveable { mutableStateOf(null as Subscreen?) }

        ModalBottomSheet(
            onDismissRequest = { listener.dismiss() },
            content = {
                BottomSheet.ListItem(
                    title = stringResource(R.string.input_config_type),
                    subtitle = when (config.type) {
                        InputConfiguration.KeyboardType.Phone -> stringResource(R.string.input_config_type_phone)
                        InputConfiguration.KeyboardType.Calculator -> stringResource(R.string.input_config_type_calculator)
                    },
                    listener = { subcreen = Subscreen.KeyboardType }
                )

                BottomSheet.ListItem(
                    title = stringResource(R.string.input_config_size),
                    subtitle = "${config.size}%",
                    listener = { listener.launchKeyboardSizeScreen(config) }
                )

                BottomSheet.ListItem(
                    title = stringResource(R.string.input_config_align_horizontal),
                    subtitle = when (config.horizontal) {
                        InputConfiguration.HorizontalAlignment.Left -> stringResource(R.string.input_config_align_left)
                        InputConfiguration.HorizontalAlignment.Center -> stringResource(R.string.input_config_align_center)
                        InputConfiguration.HorizontalAlignment.Right -> stringResource(R.string.input_config_align_right)
                    },
                    listener = { subcreen = Subscreen.HorizontalAlignment }
                )

                BottomSheet.ListItem(
                    title = stringResource(R.string.input_config_align_vertical),
                    subtitle = when (config.vertical) {
                        InputConfiguration.VerticalAlignment.Top -> stringResource(R.string.input_config_align_top)
                        InputConfiguration.VerticalAlignment.Center -> stringResource(R.string.input_config_align_center)
                        InputConfiguration.VerticalAlignment.Bottom -> stringResource(R.string.input_config_align_bottom)
                    },
                    listener = { subcreen = Subscreen.VerticalAlignment }
                )

                BottomSheet.ListItem(
                    title = stringResource(R.string.input_config_confirm_button_location),
                    subtitle = when (config.confirmButtonLocation) {
                        InputConfiguration.ConfirmButtonLocation.Left -> stringResource(R.string.input_config_align_left)
                        InputConfiguration.ConfirmButtonLocation.Right -> stringResource(R.string.input_config_align_right)
                    },
                    listener = { subcreen = Subscreen.ConfirmButtonLocation }
                )

                BottomSheet.ListItem(
                    title = stringResource(R.string.generic_save),
                    subtitle = null,
                    listener = { listener.save() }
                )
            }
        )

        when (subcreen) {
            null -> {/* nothing to show */}
            Subscreen.KeyboardType -> ModalBottomSheet(
                onDismissRequest = { subcreen = null },
                content = {
                    @Composable
                    fun Item(type: InputConfiguration.KeyboardType, label: String) {
                        BottomSheet.RadioButton(label, config.type == type) {
                            listener.adjustConfiguration { it.copy(type = type) }

                            subcreen = null
                        }
                    }

                    BottomSheet.Title(stringResource(R.string.input_config_type))

                    Item(InputConfiguration.KeyboardType.Phone, stringResource(R.string.input_config_type_phone))
                    Item(InputConfiguration.KeyboardType.Calculator, stringResource(R.string.input_config_type_calculator))
                }
            )
            Subscreen.HorizontalAlignment -> ModalBottomSheet(
                onDismissRequest = { subcreen = null },
                content = {
                    @Composable
                    fun Item(alignment: InputConfiguration.HorizontalAlignment, label: String) {
                        BottomSheet.RadioButton(label, config.horizontal == alignment) {
                            listener.adjustConfiguration { it.copy(horizontal = alignment) }

                            subcreen = null
                        }
                    }

                    BottomSheet.Title(stringResource(R.string.input_config_align_horizontal))

                    Item(InputConfiguration.HorizontalAlignment.Left, stringResource(R.string.input_config_align_left))
                    Item(InputConfiguration.HorizontalAlignment.Center, stringResource(R.string.input_config_align_center))
                    Item(InputConfiguration.HorizontalAlignment.Right, stringResource(R.string.input_config_align_right))
                }
            )
            Subscreen.VerticalAlignment -> ModalBottomSheet(
                onDismissRequest = { subcreen = null },
                content = {
                    @Composable
                    fun Item(alignment: InputConfiguration.VerticalAlignment, label: String) {
                        BottomSheet.RadioButton(label, config.vertical == alignment) {
                            listener.adjustConfiguration { it.copy(vertical = alignment) }

                            subcreen = null
                        }
                    }

                    BottomSheet.Title(stringResource(R.string.input_config_align_vertical))

                    Item(InputConfiguration.VerticalAlignment.Top, stringResource(R.string.input_config_align_top))
                    Item(InputConfiguration.VerticalAlignment.Center, stringResource(R.string.input_config_align_center))
                    Item(InputConfiguration.VerticalAlignment.Bottom, stringResource(R.string.input_config_align_bottom))
                }
            )
            Subscreen.ConfirmButtonLocation -> ModalBottomSheet(
                onDismissRequest = { subcreen = null },
                content = {
                    @Composable
                    fun Item(location: InputConfiguration.ConfirmButtonLocation, label: String) {
                        BottomSheet.RadioButton(label, config.confirmButtonLocation == location) {
                            listener.adjustConfiguration { it.copy(confirmButtonLocation = location) }

                            subcreen = null
                        }
                    }

                    BottomSheet.Title(stringResource(R.string.input_config_confirm_button_location))

                    Item(InputConfiguration.ConfirmButtonLocation.Left, stringResource(R.string.input_config_align_left))
                    Item(InputConfiguration.ConfirmButtonLocation.Right, stringResource(R.string.input_config_align_right))
                }
            )
        }
    }

    interface Listener {
        fun adjustConfiguration(updater: (InputConfiguration) -> InputConfiguration)
        fun launchKeyboardSizeScreen(configuration: InputConfiguration)
        fun save()
        fun dismiss()
    }

    private enum class Subscreen {
        KeyboardType,
        HorizontalAlignment,
        VerticalAlignment,
        ConfirmButtonLocation,
    }
}