/*
 * Calculate Android Copyright <C> 2020 - 2025 Jonas Lochmann
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation version 3 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see https://www.gnu.org/licenses/
 */
package com.rechnen.app.ui.userlist

import android.content.Intent
import android.os.Bundle
import android.text.format.DateUtils

import androidx.activity.compose.setContent
import androidx.activity.enableEdgeToEdge
import androidx.activity.result.contract.ActivityResultContracts
import androidx.activity.viewModels
import androidx.compose.foundation.combinedClickable
import androidx.compose.foundation.layout.Box
import androidx.compose.foundation.layout.Column
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.foundation.layout.fillMaxWidth
import androidx.compose.foundation.layout.padding
import androidx.compose.foundation.lazy.LazyColumn
import androidx.compose.foundation.lazy.items
import androidx.compose.material.icons.Icons
import androidx.compose.material.icons.filled.Add
import androidx.compose.material.icons.outlined.Info
import androidx.compose.material3.Card
import androidx.compose.material3.ExperimentalMaterial3Api
import androidx.compose.material3.ExtendedFloatingActionButton
import androidx.compose.material3.Icon
import androidx.compose.material3.IconButton
import androidx.compose.material3.MaterialTheme
import androidx.compose.material3.Scaffold
import androidx.compose.material3.SnackbarHost
import androidx.compose.material3.Text
import androidx.compose.material3.TopAppBar
import androidx.compose.runtime.collectAsState
import androidx.compose.runtime.livedata.observeAsState
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.platform.LocalContext
import androidx.compose.ui.res.stringResource
import androidx.compose.ui.text.style.TextAlign
import androidx.compose.ui.unit.dp
import androidx.fragment.app.FragmentActivity
import com.rechnen.app.R
import com.rechnen.app.data.modelparts.InputConfiguration
import com.rechnen.app.data.modelparts.TaskDifficulty

import com.rechnen.app.ui.AboutActivity
import com.rechnen.app.ui.AppTheme
import com.rechnen.app.ui.training.TrainingMainActivity
import com.rechnen.app.ui.training.TrainingMode
import com.rechnen.app.ui.user.ConfigureScaleFactorActivity
import com.rechnen.app.ui.user.CreateUserDialog
import com.rechnen.app.ui.user.CreateUserDialogListener
import com.rechnen.app.ui.user.DeleteUserDialog
import com.rechnen.app.ui.user.DeleteUserDialogListener
import com.rechnen.app.ui.user.EditUserDialog
import com.rechnen.app.ui.user.EditUserDifficultyDialog
import com.rechnen.app.ui.user.EditUserInputConfigurationDialog
import com.rechnen.app.ui.user.EditUserListener
import com.rechnen.app.ui.user.ViewUserStatisticDialog
import com.rechnen.app.ui.util.Time

class SelectUserActivity : FragmentActivity() {
    private val model by viewModels<SelectUserModel>()

    val adjustScaleFactor = registerForActivityResult(
        ActivityResultContracts.StartActivityForResult()
    ) { result ->
        if (result.resultCode == RESULT_OK) {
            model.adjustInput {
                it.copy(
                    size = result.data!!.getIntExtra(ConfigureScaleFactorActivity.RESULT_EXTRA_FACTOR, -1)
                )
            }
        }
    }

    @OptIn(ExperimentalMaterial3Api::class)
    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)

        enableEdgeToEdge()

        setContent {
            val users = model.users.observeAsState()
            val dialog = model.dialog.collectAsState()

            AppTheme {
                Scaffold(
                    topBar = {
                        TopAppBar(
                            title = { Text(stringResource(R.string.app_name)) },
                            actions = {
                                IconButton(
                                    onClick = { showAboutScreen() },
                                    content = { Icon(
                                        Icons.Outlined.Info,
                                        stringResource(R.string.about_title)
                                    ) }
                                )
                            }
                        )
                    },
                    floatingActionButton = {
                        ExtendedFloatingActionButton(
                            onClick = { model.createUser() },
                            icon = { Icon(Icons.Default.Add, stringResource(R.string.user_list_create_user)) },
                            text = { Text(stringResource(R.string.user_list_create_user)) }
                        )
                    },
                    snackbarHost = { SnackbarHost(model.snackbar) },
                    content = { insets ->
                        val currentUsers = users.value

                        if (currentUsers == null) {/* nothing to show */}
                        else if (currentUsers.isEmpty()) {
                            Box (Modifier.fillMaxSize().padding(insets).padding(16.dp)) {
                                Text(
                                    stringResource(R.string.user_list_empty_text),
                                    textAlign = TextAlign.Center,
                                    modifier = Modifier.fillMaxWidth().align(Alignment.CenterStart)
                                )
                            }
                        } else {
                            LazyColumn(
                                modifier = Modifier.fillMaxSize(),
                                contentPadding = insets,
                                content = {
                                    items(currentUsers, key = { it.id }) { user ->
                                        Card (
                                            Modifier
                                                .animateItem()
                                                .padding(horizontal = 8.dp, vertical = 4.dp)
                                                .combinedClickable(
                                                    enabled = true,
                                                    onClick = { startTraining(user.id, TrainingMode.Regular) },
                                                    onLongClick = { model.editUser(user) }
                                                )
                                        ) {
                                            Column (
                                                Modifier
                                                    .padding(8.dp)
                                                    .fillMaxWidth()
                                            ) {
                                                Text(user.name, style = MaterialTheme.typography.titleLarge)

                                                if (user.playTime > 0) {
                                                    Text(stringResource(
                                                        R.string.user_list_playtime,
                                                        Time.formatTime(user.playTime, LocalContext.current)
                                                    ))
                                                }

                                                Text(
                                                    if (user.lastUse == null) stringResource(R.string.user_list_statistic_empty)
                                                    else stringResource(
                                                        R.string.user_list_statistic,
                                                        DateUtils.formatDateTime(LocalContext.current, user.lastUse, DateUtils.FORMAT_SHOW_DATE or DateUtils.FORMAT_SHOW_TIME),
                                                        resources.getQuantityString(R.plurals.user_list_blocks, user.lastBlockCount, user.lastBlockCount)
                                                    )
                                                )
                                            }
                                        }
                                    }
                                }
                            )
                        }
                    }
                )

                when (val dialog = dialog.value) {
                    is SelectUserModel.Dialog.None -> {/* nothing to do */}
                    is SelectUserModel.Dialog.CreateUser -> CreateUserDialog(
                        object: CreateUserDialogListener {
                            override fun create(name: String) {
                                model.createUser(name)
                            }

                            override fun dismiss() {
                                model.dismiss(dialog)
                            }
                        }
                    )
                    is SelectUserModel.Dialog.EditUser -> EditUserDialog(
                        dialog.user,
                        object: EditUserListener {
                            override fun startTimeTrial() {
                                startTraining(dialog.user.id, TrainingMode.TimeTrail)

                                model.dismiss(dialog)
                            }

                            override fun viewStatistic() {
                                model.showStatistic(dialog.user)
                            }

                            override fun configureDifficulty() {
                                model.configureDifficulty(dialog.user)
                            }

                            override fun configureInput() {
                                model.configureInput(dialog.user)
                            }

                            override fun deleteUser() {
                                model.deleteUser(dialog.user)
                            }

                            override fun dismiss() {
                                model.dismiss(dialog)
                            }
                        }
                    )
                    is SelectUserModel.Dialog.UserStatistic -> ViewUserStatisticDialog(
                        dialog.user, dialog.stats
                    ) { model.dismiss(dialog) }
                    is SelectUserModel.Dialog.ConfigureDifficulty -> EditUserDifficultyDialog.Main(
                        dialog.newDifficulty,
                        object: EditUserDifficultyDialog.Listener {
                            override fun adjustDifficulty(updater: (TaskDifficulty) -> TaskDifficulty) {
                                model.adjustDifficulty(updater)
                            }

                            override fun save() {
                                model.saveDifficulty(dialog)
                            }

                            override fun dismiss() {
                                model.dismiss(dialog)
                            }
                        }
                    )
                    is SelectUserModel.Dialog.ConfigureInput -> EditUserInputConfigurationDialog.Main(
                        dialog.newConfig,
                        object: EditUserInputConfigurationDialog.Listener {
                            override fun adjustConfiguration(updater: (InputConfiguration) -> InputConfiguration) {
                                model.adjustInput(updater)
                            }

                            override fun launchKeyboardSizeScreen(configuration: InputConfiguration) {
                                adjustScaleFactor.launch(ConfigureScaleFactorActivity.buildIntent(this@SelectUserActivity, configuration))
                            }

                            override fun save() {
                                model.saveInput(dialog)
                            }

                            override fun dismiss() {
                                model.dismiss(dialog)
                            }
                        }
                    )
                    is SelectUserModel.Dialog.DeleteUser -> DeleteUserDialog(
                        user = dialog.user,
                        listener = object: DeleteUserDialogListener {
                            override fun confirm() {
                                model.confirmDeleteUser(dialog.user)
                            }

                            override fun cancel() {
                                model.dismiss(dialog)
                            }
                        }
                    )
                }
            }
        }
    }

    private fun showAboutScreen() {
        startActivity(Intent(this, AboutActivity::class.java))
    }

    private fun startTraining(userId: Int, trainingMode: TrainingMode) {
        TrainingMainActivity.launch(this@SelectUserActivity, userId, trainingMode)
    }
}
