/*
 * Calculate Android Copyright <C> 2020 - 2025 Jonas Lochmann
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation version 3 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see https://www.gnu.org/licenses/
 */
package com.rechnen.app.ui.userlist

import android.app.Application
import android.util.Log
import androidx.compose.material3.SnackbarHostState
import androidx.lifecycle.AndroidViewModel
import androidx.lifecycle.viewModelScope
import com.rechnen.app.BuildConfig
import com.rechnen.app.R
import com.rechnen.app.data.AppDatabase
import com.rechnen.app.data.model.AggregatedStatisticTaskResult
import com.rechnen.app.data.model.User
import com.rechnen.app.data.modelparts.InputConfiguration
import com.rechnen.app.data.modelparts.TaskDifficulty
import com.rechnen.app.data.modelparts.TaskDifficultyProfiles
import kotlinx.coroutines.Dispatchers
import kotlinx.coroutines.flow.MutableStateFlow
import kotlinx.coroutines.flow.StateFlow
import kotlinx.coroutines.flow.getAndUpdate
import kotlinx.coroutines.flow.update
import kotlinx.coroutines.invoke
import kotlinx.coroutines.launch

class SelectUserModel(private val application: Application): AndroidViewModel(application) {
    companion object {
        private const val LOG_TAG = "SelectUserModel"
    }

    private val database = AppDatabase.with(application)
    private val dialogInternal = MutableStateFlow<Dialog>(Dialog.None)

    val users = database.user().getAllUsers()
    val dialog: StateFlow<Dialog> = dialogInternal
    val snackbar = SnackbarHostState()

    fun createUser() {
        dialogInternal.compareAndSet(Dialog.None, Dialog.CreateUser)
    }

    fun createUser(name: String) {
        if (!dialogInternal.compareAndSet(Dialog.CreateUser, Dialog.None)) return

        viewModelScope.launch {
            if (name.isBlank()) {
                snackbar.showSnackbar(application.getString(R.string.user_list_create_user_no_name_error))
            } else {
                val ok = try {
                    Dispatchers.IO.invoke {
                        database.user().addUserSync(User(
                            id = 0,
                            name = name,
                            difficulty = TaskDifficultyProfiles.medium,
                            lastBlockCount = 0,
                            lastUse = null,
                            playTime = 0,
                            solvedTasks = 0
                        ))
                    }

                    true
                } catch (ex: Exception) {
                    if (BuildConfig.DEBUG) {
                        Log.d(LOG_TAG, "could not create user", ex)
                    }

                    snackbar.showSnackbar(application.getString(R.string.generic_error))

                    false
                }

                if (ok) snackbar.showSnackbar(application.getString(R.string.user_list_created_user, name))
            }
        }
    }

    fun editUser(user: User) {
        dialogInternal.update {
            if (it == Dialog.None) Dialog.EditUser(user)
            else it
        }
    }

    fun showStatistic(user: User) {
        viewModelScope.launch {
            val stats = database.statisticTaskResult().queryAggregatedCoroutine(user.id)

            dialogInternal.update {
                if (it == Dialog.EditUser(user)) Dialog.UserStatistic(user, stats)
                else it
            }
        }
    }

    fun configureDifficulty(user: User) {
        dialogInternal.update {
            if (it == Dialog.EditUser(user)) Dialog.ConfigureDifficulty(user, user.difficulty)
            else it
        }
    }

    fun adjustDifficulty(updater: (TaskDifficulty) -> TaskDifficulty) {
        dialogInternal.update {
            if (it is Dialog.ConfigureDifficulty) it.copy(newDifficulty = updater(it.newDifficulty))
            else it
        }
    }

    fun saveDifficulty(dialog: Dialog.ConfigureDifficulty) {
        if (!dialogInternal.compareAndSet(dialog, Dialog.None)) return

        viewModelScope.launch {
            if (dialog.user.difficulty == dialog.newDifficulty) {
                snackbar.showSnackbar(application.getString(R.string.difficulty_not_modified_toast))
            } else {
                val ok = try {
                    Dispatchers.IO.invoke {
                        database.runInTransaction {
                            database.savedTask().deleteUserTasksSync(dialog.user.id)
                            database.statisticTaskResult().deleteStatisticTaskResultsForUserSync(dialog.user.id)
                            database.user().updateUserDifficultyConfigSync(dialog.user.id, dialog.newDifficulty)
                        }
                    }

                    true
                } catch (ex: Exception) {
                    if (BuildConfig.DEBUG) {
                        Log.d(LOG_TAG, "could not update user", ex)
                    }

                    snackbar.showSnackbar(application.getString(R.string.generic_error))

                    false
                }

                if (ok) snackbar.showSnackbar(application.getString(R.string.difficulty_save_toast))
            }
        }
    }

    fun configureInput(user: User) {
        dialogInternal.update {
            if (it == Dialog.EditUser(user)) Dialog.ConfigureInput(user, user.inputConfiguration)
            else it
        }
    }

    fun adjustInput(updater: (InputConfiguration) -> InputConfiguration) {
        dialogInternal.update {
            if (it is Dialog.ConfigureInput) it.copy(newConfig = updater(it.newConfig))
            else it
        }
    }

    fun saveInput(dialog: Dialog.ConfigureInput) {
        if (!dialogInternal.compareAndSet(dialog, Dialog.None)) return

        viewModelScope.launch {
            if (dialog.user.inputConfiguration == dialog.newConfig) {
                snackbar.showSnackbar(application.getString(R.string.difficulty_not_modified_toast))
            } else {
                val ok = try {
                    Dispatchers.IO.invoke {
                        database.user().updateUserInputConfigurationSync(dialog.user.id, dialog.newConfig)
                    }

                    true
                } catch (ex: Exception) {
                    if (BuildConfig.DEBUG) {
                        Log.d(LOG_TAG, "could not update user", ex)
                    }

                    snackbar.showSnackbar(application.getString(R.string.generic_error))

                    false
                }

                if (ok) snackbar.showSnackbar(application.getString(R.string.difficulty_save_toast))
            }
        }
    }

    fun deleteUser(user: User) {
        dialogInternal.update {
            if (it == Dialog.EditUser(user)) Dialog.DeleteUser(user)
            else it
        }
    }

    fun dismiss(dialog: Dialog) {
        dialogInternal.update {
            if (it == dialog) Dialog.None
            else it
        }
    }

    fun confirmDeleteUser(user: User) {
        val old = dialogInternal.getAndUpdate {
            if (it == Dialog.DeleteUser(user)) Dialog.None
            else it
        }

        if (old == Dialog.DeleteUser(user)) {
            viewModelScope.launch {
                val ok = try {
                    Dispatchers.IO.invoke {
                        database.user().deleteUserByIdSync(user.id)
                    }

                    true
                } catch (ex: Exception) {
                    if (BuildConfig.DEBUG) {
                        Log.d(LOG_TAG, "could not remove user", ex)
                    }

                    snackbar.showSnackbar(application.getString(R.string.generic_error))

                    false
                }

                if (ok) {
                    snackbar.showSnackbar(application.getString(R.string.user_list_deleted_user))
                }
            }
        }
    }

    sealed class Dialog {
        object None: Dialog()
        object CreateUser: Dialog()
        data class EditUser(val user: User): Dialog()
        data class UserStatistic(val user: User, val stats: List<AggregatedStatisticTaskResult>): Dialog()
        data class ConfigureDifficulty(val user: User, val newDifficulty: TaskDifficulty): Dialog()
        data class ConfigureInput(val user: User, val newConfig: InputConfiguration): Dialog()
        data class DeleteUser(val user: User): Dialog()
    }
}